/*
 * EMail extension to Oracle RDBMS
 */
#include <stdlib.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <oci.h>

#ifndef SEND_PROG
#define SEND_PROG "/var/qmail/bin/qmail-inject"
#endif

#ifndef SEND_OPTS
#define SEND_OPTS "-H"
#endif

#ifndef FROM_OPT
#define FROM_OPT "-f"
#endif

#ifndef MAXBUFLEN
#define MAXBUFLEN 1024
#endif

/* do something about later... */
static void handle_error(int out, OCIExtProcContext *with_context,int error,OCIError *err)

{
    char msg[2000];
    int i=1;
    ub4 errcode;
    switch(error){
    case OCI_SUCCESS:
    case OCI_SUCCESS_WITH_INFO:
	strncpy(msg,"X-OraError: no error\n",sizeof(msg));
	break;
    case OCI_NEED_DATA:
	strncpy(msg,"X-OraError: need data\n",sizeof(msg));
	break;
    case OCI_INVALID_HANDLE:
	strncpy(msg,"X-OraError: Invalid handle\n",sizeof(msg));
	break;
    case OCI_ERROR:
	{
	    char buf[2000];
    	    OCIErrorGet(err,i++,0,&errcode,buf,sizeof(buf),OCI_HTYPE_ERROR);
    	    OCIExtProcRaiseExcpWithMsg(with_context,errcode,msg, strlen(msg));
	    sprintf(msg,"X-OraError: Error ORA-%i %s\n",errcode,buf);
	    break;
	}
    }
    write(out,msg,strlen(msg));
}

static void smail(OCIExtProcContext *with_context,int in, int out, const char* send_to, const char* sent_from)
{
    close(out);
    if(dup2(in,0)==-1)
    {
	char msg[2000];
	sprintf(msg,"Error forking mail program: OS error: %s",strerror(errno));
	OCIExtProcRaiseExcpWithMsg(with_context,20500,msg, strlen(msg));
	return;
    }
    close(in);
    fcntl(0,F_SETFD,0);
    execl(SEND_PROG, SEND_PROG, SEND_OPTS, FROM_OPT, sent_from, send_to, NULL);
    /* Can't happen */
    {
	char msg[2000];
	sprintf(msg,"Error sending mail: OS error: %s",strerror(errno));
	OCIExtProcRaiseExcpWithMsg(with_context,20500,msg, strlen(msg));
    }
}

static void compose_message(OCIExtProcContext *with_context,int out,OCILobLocator *headers, OCILobLocator* message)
{
    ub4 amt=0;
    ub4 offset=1;
    sword retval;
    ub1 buf[MAXBUFLEN];
    OCIEnv *env;
    OCISvcCtx *svctx;
    OCIError *err;
    OCIExtProcGetEnv(with_context,&env,&svctx,&err);
#define IV_STDXHEADERS "X-Mailer: ivOraMail\n"
    write(out, IV_STDXHEADERS,sizeof(IV_STDXHEADERS)-1);
    do {
	memset(buf,0,sizeof(buf));
	amt=0;
	retval=OCILobRead(svctx,err,headers,&amt,offset,buf,
			  sizeof(buf),0,0,0,SQLCS_IMPLICIT);
	if(retval!=OCI_NEED_DATA&&retval!=OCI_SUCCESS){
	    /* serious error... */
	    handle_error(out,with_context,retval,err);
	    return;
	}
	write(out,buf,strlen(buf));
    } while(retval!=OCI_SUCCESS);
    write(out,"\n",1);/*put extra carriege return at the end to separate mime headers*/
    do {
        memset(buf,0,sizeof(buf));
        amt=0;//xFFFFFFFF; /* Max LOB size */
	retval=OCILobRead(svctx,err,message,&amt,offset,buf,
			  sizeof(buf),0,0,0,SQLCS_IMPLICIT);
	if(retval!=OCI_NEED_DATA&&retval!=OCI_SUCCESS){
	    handle_error(out,with_context,retval,err);
	    break;
	}
	//offset+=amt;
	write(out,buf,strlen(buf));
    } while(retval!=OCI_SUCCESS);
}

int DoSendMail(
		OCIExtProcContext *with_context,
		char *send_to,
		char *sent_from,
		OCILobLocator *headers,
		int hdr_ind,
		OCILobLocator *message,
		int msg_ind
	)
{
    char cmd[1024];
    int mail[2];
    int mail_pid;
    if(pipe(mail)==-1){
	char msg[2000];
	sprintf(msg,"Error sending mail: OS error: %s",strerror(errno));
	OCIExtProcRaiseExcpWithMsg(with_context,20500,msg, strlen(msg));
	return 0;
    }
    mail_pid=fork();
    if(mail_pid==-1){
	char msg[2000];
	sprintf(msg,"Error sending mail: OS error: %s",strerror(errno));
	OCIExtProcRaiseExcpWithMsg(with_context,20500,msg, strlen(msg));
	return 0;
    }
    if(mail_pid==0){
	/* If this returns, then there was an error. Exception was rised, and
	   we should stop doing whatever we are doing */
	smail(with_context,mail[0],mail[1],send_to,sent_from);
	exit(0);
    }
    close(mail[0]);
    compose_message(with_context,mail[1],headers,message);
    close(mail[1]);
    waitpid(mail_pid,NULL,0);
    return 1;
}
